from typing import Tuple

from .database import connect_to_uri
from .diff_tables import TableSegment, TableDiffer

def create_source(db_uri: str, table_name: str, key_column: str, extra_columns: Tuple[str, ...] = ()):
    db = connect_to_uri(db_uri)
    return TableSegment(db, (table_name,), key_column, tuple(extra_columns))


"""Provides classes for a pseudo-SQL AST that compiles to SQL code
"""

from typing import List, Union, Tuple, Optional

from runtype import dataclass

DbPath = Tuple[str, ...]
DbKey = Union[int, str, bytes]

class Sql:
    pass

SqlOrStr = Union[Sql, str]

@dataclass
class Compiler:
    """Provides a set of utility methods for compiling SQL

    For internal use.
    """

    database: object #Database
    in_select: bool = False     # Compilation

    def quote(self, s: str):
        return self.database.quote(s)

    def compile(self, elem):
        if isinstance(elem, Sql):
            return elem.compile(self)
        elif isinstance(elem, str):
            return elem
        elif isinstance(elem, int):
            return str(elem)
        assert False

@dataclass
class TableName(Sql):
    name: DbPath

    def compile(self, c: Compiler):
        return c.quote('.'.join(self.name))

@dataclass
class Value(Sql):
    value: object   # Primitive

    def compile(self, c: Compiler):
        if isinstance(self.value, bytes):
            return "b'%s'" % self.value.decode()
        elif isinstance(self.value, str):
            return "'%s'" % self.value
        return str(self.value)

@dataclass
class Select(Sql):
    columns: List[SqlOrStr]
    table: SqlOrStr = None
    where: List[SqlOrStr] = None
    order_by: List[SqlOrStr] = None
    group_by: List[SqlOrStr] = None

    def compile(self, parent_c: Compiler):
        c = parent_c.replace(in_select=True)
        columns = ', '.join(map(c.compile, self.columns))
        select = f'SELECT {columns}'

        if self.table:
            select += ' FROM ' + c.compile(self.table)
        
        if self.where:
            select += ' WHERE ' + ' AND '.join(map(c.compile, self.where))

        if self.group_by:
            select += ' GROUP BY ' + ', '.join(map(c.compile, self.group_by))

        if self.order_by:
            select += ' ORDER BY ' + ', '.join(map(c.compile, self.order_by))

        if parent_c.in_select:
            select = '(%s)' % select
        return select

@dataclass
class Enum(Sql):
    table: DbPath
    order_by: SqlOrStr

    def compile(self, c: Compiler):
        table = c.quote('.'.join(self.table))
        order = c.compile(self.order_by)
        return f'(SELECT *, (row_number() over (ORDER BY {order})) as idx FROM {table} ORDER BY {order}) tmp'



@dataclass
class Checksum(Sql):
    exprs: List[SqlOrStr]

    def compile(self, c: Compiler):
        compiled_exprs = ', '.join(c.database.to_string(s) for s in map(c.compile, self.exprs))
        expr =  f'concat({compiled_exprs})'
        md5 = c.database.md5_to_int(expr)
        return f'sum({md5})'

@dataclass
class Compare(Sql):
    op: str
    a: SqlOrStr
    b: SqlOrStr

    def compile(self, c: Compiler):
        return f'({c.compile(self.a)} {self.op} {c.compile(self.b)})'

@dataclass
class In(Sql):
    expr: SqlOrStr
    list: List #List[SqlOrStr]

    def compile(self, c: Compiler):
        elems = ', '.join(map(c.compile, self.list))
        return f'({c.compile(self.expr)} IN ({elems}))'

@dataclass
class Count(Sql):
    column: Optional[SqlOrStr] = None

    def compile(self, c: Compiler):
        if self.column:
            return f"count({c.compile(self.column)})"
        return 'count(*)'



import sys
import time
import logging
from itertools import islice

from .diff_tables import TableSegment, TableDiffer
from .database import connect_to_uri

import click

LOG_FORMAT = '[%(asctime)s] %(levelname)s - %(message)s'
DATE_FORMAT ='%H:%M:%S'

@click.command()
@click.argument('db1_uri')
@click.argument('table1_name')
@click.argument('db2_uri')
@click.argument('table2_name')
@click.option('-k', '--key_column', default='id', help='Name of primary key column')
@click.option('-c', '--columns', default=['updated_at'], multiple=True, help='Names of extra columns to compare')
@click.option('-l', '--limit', default=None, help='Maximum number of differences to find')
@click.option('--bisection-factor', default=32, help='Segments per iteration')
@click.option('--bisection-threshold', default=1024**2, help='Minimal bisection threshold')
@click.option('-s', '--stats', is_flag=True, help='Print stats instead of a detailed diff')
@click.option('-d', '--debug', is_flag=True, help='Print debug info')
@click.option('-v', '--verbose', is_flag=True, help='Print extra info')
def main(db1_uri, table1_name, db2_uri, table2_name, key_column, columns, limit, bisection_factor, bisection_threshold, stats, debug, verbose):
    if limit and stats:
        print("Error: cannot specify a limit when using the -s/--stats switch")
        return

    if debug:
        logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT, datefmt=DATE_FORMAT)
    elif verbose:
        logging.basicConfig(level=logging.INFO, format=LOG_FORMAT, datefmt=DATE_FORMAT)


    db1 = connect_to_uri(db1_uri)
    db2 = connect_to_uri(db2_uri)

    start = time.time()

    table1 = TableSegment(db1, (table1_name,), key_column, columns)
    table2 = TableSegment(db2, (table2_name,), key_column, columns)

    differ = TableDiffer(bisection_factor=bisection_factor, bisection_threshold=bisection_threshold, debug=debug)
    diff_iter = differ.diff_tables(table1, table2)

    if limit:
        diff_iter = islice(diff_iter, int(limit))

    if stats:
        diff = list(diff_iter)
        percent = 100 * len(diff) / table1.count
        print(f"Diff-Total: {len(diff)} changed rows out of {table1.count}")
        print(f"Diff-Percent: {percent:.4f}%")
        plus = len([1 for op,_ in diff if op=='+'])
        minus = len([1 for op,_ in diff if op=='-'])
        print(f"Diff-Split: +{plus}  -{minus}")
    else:
        for op, key in diff_iter:
            print(op, key)
            sys.stdout.flush()

    end = time.time()

    logging.info(f"Duration: {end-start:.2f} seconds.")


if __name__ == '__main__':
    main()


from abc import ABC, abstractmethod
import logging
from typing import Tuple

import dsnparse

from .sql import SqlOrStr, Compiler


logger = logging.getLogger('database')


def import_postgres():
    import psycopg2
    import psycopg2.extras
    psycopg2.extensions.set_wait_callback(psycopg2.extras.wait_select)
    return psycopg2

def import_mysql():
    import mysql.connector
    return mysql.connector

def import_snowflake():
    import snowflake.connector
    return snowflake

def import_mssql():
    import pymssql
    return pymssql

def import_oracle():
    import cx_Oracle
    return cx_Oracle


class ConnectError(Exception):
    pass


def _one(seq):
    x ,= seq
    return x

class Database(ABC):
    """Base abstract class for databases.
    
    Used for providing connection code and implementation specific SQL utilities.
    """

    def _query(self, sql_code: str) -> list:
        "Uses the standard SQL cursor interface"
        c = self._conn.cursor()
        c.execute(sql_code)
        return c.fetchall()

    def query(self, sql_ast: SqlOrStr, res_type: type):
        "Query the given SQL AST, and attempt to convert the result to type 'res_type'"
        sql_code = Compiler(self).compile(sql_ast)
        logger.debug("Running SQL (%s): %s", type(self).__name__, sql_code)
        res = self._query(sql_code)
        if res_type is int:
            res = _one(_one(res))
            if res is None:     # May happen due to sum() of 0 items
                return None
            return int(res)
        elif getattr(res_type, '__origin__', None) is list and len(res_type.__args__) == 1:
            if res_type.__args__ == (int,):
                return [_one(row) for row in res]
            elif res_type.__args__ == (Tuple,):
                return res
            else:
                raise ValueError(res_type)
        return res

    @abstractmethod
    def quote(self, s: str):
        "Quote SQL name (implementation specific)"
        ...

    @abstractmethod
    def to_string(self, s: str) -> str:
        "Provide SQL for casting a column to string"
        ...

    @abstractmethod
    def md5_to_int(self, s: str) -> str:
        "Provide SQL for computing md5 and returning an int"
        ...

CHECKSUM_HEXDIGITS = 15     # Must be 15 or lower
MD5_HEXDIGITS = 32

_CHECKSUM_BITSIZE = CHECKSUM_HEXDIGITS<<2
CHECKSUM_MASK = (2**_CHECKSUM_BITSIZE) - 1

class Postgres(Database):
    def __init__(self, host, port, database, user, password):
        postgres = import_postgres()
        self.args = dict(host=host, port=port, database=database, user=user, password=password)

        try:
            self._conn = postgres.connect(**self.args)
        except postgres.OperationalError as e:
            raise ConnectError(*e.args) from e

    def quote(self, s: str):
        return f'"{s}"'

    def md5_to_int(self, s: str) -> str:
        return f"('x' || substring(md5({s}), {1+MD5_HEXDIGITS-CHECKSUM_HEXDIGITS}))::bit({_CHECKSUM_BITSIZE})::bigint"

    def to_string(self, s: str):
        return f'{s}::varchar'


class MySQL(Database):
    def __init__(self, host, port, database, user, password):
        mysql = import_mysql()

        args = dict(host=host, port=port, database=database, user=user, password=password)
        self._args = {k:v for k, v in args.items() if v is not None}

        try:
            self._conn = mysql.connect(charset='utf8', use_unicode=True, **self._args)
        except mysql.Error as e:
            if e.errno == mysql.errorcode.ER_ACCESS_DENIED_ERROR:
                raise ConnectError("Bad user name or password") from e
            elif e.errno == mysql.errorcode.ER_BAD_DB_ERROR:
                raise ConnectError("Database does not exist") from e
            else:
                raise ConnectError(*e.args) from e

    def quote(self, s: str):
        return f'`{s}`'

    def md5_to_int(self, s: str) -> str:
        return f"cast(conv(substring(md5({s}), {1+MD5_HEXDIGITS-CHECKSUM_HEXDIGITS}), 16, 10) as unsigned)"

    def to_string(self, s: str):
        return f'cast({s} as char)' 


class Oracle(Database):
    def __init__(self, host, port, database, user, password):
        assert not port
        oracle = import_oracle()
        try:
            self._conn = oracle.connect(user=user, password=password, dsn="%s/%s" % (host, database))
        except Exception as e:
            raise ConnectError(*e.args) from e

    def md5_to_int(self, s: str) -> str:
        # standard_hash is faster than DBMS_CRYPTO.Hash
        # TODO: Find a way to use UTL_RAW.CAST_TO_BINARY_INTEGER ?
        return f"to_number(substr(standard_hash({s}, 'MD5'), 18), 'xxxxxxxxxxxxxxx')"

    def quote(self, s: str):
        return f'{s}'

    def to_string(self, s: str):
        return f'cast({s} as varchar(1024))' 

class Redshift(Postgres):
    def md5_to_int(self, s: str) -> str:
        return f"strtol(substring(md5({s}), {1+MD5_HEXDIGITS-CHECKSUM_HEXDIGITS}), 16)::decimal(38)"

class MsSQL(Database):
    "AKA sql-server"

    def __init__(self, host, port, database, user, password):
        mssql = import_mssql()

        args = dict(server=host, port=port, database=database, user=user, password=password)
        self._args = {k:v for k, v in args.items() if v is not None}

        try:
            self._conn = mssql.connect(**self._args)
        except mssql.Error as e:
            raise ConnectError(*e.args) from e

    def quote(self, s: str):
        return f'[{s}]'

    def md5_to_int(self, s: str) -> str:
        return f"CONVERT(decimal(38,0), CONVERT(bigint, HashBytes('MD5', {s}), 2))"
        # return f"CONVERT(bigint, (CHECKSUM({s})))"

    def to_string(self, s: str):
        return f"CONVERT(varchar, {s})"

class BigQuery(Database):
    def __init__(self, project, dataset):
        from google.cloud import bigquery
        self._client = bigquery.Client(project)

    def quote(self, s: str):
        return f'`{s}`'

    def md5_to_int(self, s: str) -> str:
        return f"cast(cast( ('0x' || substr(TO_HEX(md5({s})), 18)) as int64) as numeric)"

    def _canonize_value(self, value):
        if isinstance(value, bytes):
            return value.decode()
        return value

    def _query(self, sql_code: str):
        from google.cloud import bigquery
        try:
            res = list(self._client.query(sql_code))
        except Exception as e:
            msg = "Exception when trying to execute SQL code:\n    %s\n\nGot error: %s"
            raise ConnectError(msg%(sql_code, e))

        if res and isinstance(res[0], bigquery.table.Row):
            res = [tuple(self._canonize_value(v) for v in row.values()) for row in res]
        return res

    def to_string(self, s: str):
        return f'cast({s} as string)' 

class Snowflake(Database):
    def __init__(self, account, user, password, path, schema, database, print_sql=False):
        snowflake = import_snowflake()
        logging.getLogger('snowflake.connector').setLevel(logging.WARNING)

        self._conn = snowflake.connector.connect(
            user=user,
            password=password,
            account=account
            )
        self._conn.cursor().execute(f"USE WAREHOUSE {path.lstrip('/')}")
        self._conn.cursor().execute(f"USE DATABASE {database}")
        self._conn.cursor().execute(f"USE SCHEMA {schema}")

    def quote(self, s: str):
        return s

    def md5_to_int(self, s: str) -> str:
        return f"BITAND(md5_number_lower64({s}), {CHECKSUM_MASK})"

    def to_string(self, s: str):
        return f'cast({s} as string)' 


def connect_to_uri(db_uri: str) -> Database:
    """Connect to the given database uri

    Supported databases:
    - postgres
    - mysql
    - mssql
    - oracle
    - snowflake
    - bigquery
    - redshift
    """

    dsn = dsnparse.parse(db_uri)
    if len(dsn.schemes) > 1:
        raise NotImplementedError("No support for multiple schemes")
    scheme ,= dsn.schemes

    if len(dsn.paths) == 0:
        path = ''
    elif len(dsn.paths) == 1:
        path ,= dsn.paths
    else:
        raise ValueError("Bad value for uri, too many paths: %s" % db_uri)

    if scheme == 'postgres':
        return Postgres(dsn.host, dsn.port, path, dsn.user, dsn.password)
    elif scheme == 'mysql':
        return MySQL(dsn.host, dsn.port, path, dsn.user, dsn.password)
    elif scheme == 'snowflake':
        return Snowflake(dsn.host, dsn.user, dsn.password, path, **dsn.query)
    elif scheme == 'mssql':
        return MsSQL(dsn.host, dsn.port, path, dsn.user, dsn.password)
    elif scheme == 'bigquery':
        return BigQuery(dsn.host, path)
    elif scheme == 'redshift':
        return Redshift(dsn.host, dsn.port, path, dsn.user, dsn.password)
    elif scheme == 'oracle':
        return Oracle(dsn.host, dsn.port, path, dsn.user, dsn.password)

    raise NotImplementedError(f"Scheme {dsn.scheme} currently not supported")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['data_diff']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1,<9.0', 'dsnparse', 'pyparsing', 'runtype>=0.2.4,<0.3.0']

extras_require = \
{'mysql': ['mysql-connector-python'],
 'pgsql': ['psycopg2'],
 'preql': ['preql>=0.2.11,<0.3.0'],
 'snowflake': ['snowflake-connector-python']}

entry_points = \
{'console_scripts': ['data_diff = data_diff.__main__:main']}

setup_kwargs = {
    'name': 'data-diff',
    'version': '0.0.3',
    'description': 'A cross-database, efficient diff between mostly-similar database tables',
    'long_description': '# Data Diff\n\nA cross-database, efficient diff between mostly-similar database tables.\n\nUse cases:\n\n- Quickly validate that a table was copied correctly\n\n- Find changes between two versions of the same table\n\nWe currently support the following databases:\n\n- PostgreSQL\n\n- MySQL\n\n- Oracle\n\n- Snowflake\n\n- BigQuery\n\n- Redshift\n\n\n# How does it work?\n\nData Diff finds the differences between two tables by utilizing checksum calculations and logarithmic search.\n\nInstead of comparing the entire table, it compares the tuple (primary_key, version_column), where the primary key is a unique identifier of the rows, and the version_column updates each time the row changes to a new value, that is unique to that update. Usually the versioning column would be a timestamp like `updated_at`, that would auto-update by the database. But it could also be an auto-counting integer, and so on.\n\nData Diff runs a checksum on these columns using MD5. If the checksums are not the same, we know the tables are different. We then split each table into "n" different segments of similar size (determined by the bisection factor), and repeat the comparison for each matching pair of segments. When segments are below a certain size (bisection threshold), we instead download the segments to the client, and diff them locally.\n\nData Diff splits the segments using "checkpoints", to ensure that inserted or deleted rows don\'t affect the quality of the diff.\n\nThis process is incremental, so differences are printed to stdout as they are found. Users can ensure Data Diff quits after finding some number of differences, either by providing the `--limit` option, or by closing the pipe (for example, by piping to `head`).\n\nThe algorithm goes like this:\n\n0. Table segments `A` and `B` are set to the two tables for comparison.\n\n1. Calculate the checksums on `A` and `B` using MD5.\n\n    1. If they are the same, the tables are considered equal. Stop.\n\n    2. If their size is below the threshold, diff them locallly and print the results.\n\n    3. Else:  (they are different and above the threshold)\n\n        1. Select `n-1` rows (checkpoints) in table `A`, splitting it into `n` segments of similar size.\n\n        2. Filter out checkpoints that don\'t exist in table `B`.\n\n        3. Split both `A` and `B` into `m <= n` segments according to the mutual checkpoints. `m` must be at least 2.\n\n        4. For each pair of segments `Ai` and `Bi` (where `0 <= i <= m`), recurse into step 1.\n\n## Example\n\nThe following printout shows the diff of two tables, Original and Original_1diff, with 25 million rows each, and just 1 different row between them.\n\nWe ran it with a very low bisection factor, and with the verbose flag, to demonstrate how it works.\n\nNote: It\'s usually much faster to use high bisection factors, especially when there are very few changes, like in this example.\n\n```python\n$ data_diff postgres:/// Original  postgres:/// Original_1diff  -v --bisection-factor=4\n[16:55:19] INFO - Diffing tables of size 25000095 and 25000095 | segments: 4, bisection threshold: 1048576.\n[16:55:36] INFO - Diffing segment 0/4 of size 8333364 and 8333364\n[16:55:45] INFO - . Diffing segment 0/4 of size 2777787 and 2777787\n[16:55:52] INFO - . . Diffing segment 0/4 of size 925928 and 925928\n[16:55:54] INFO - . . . Diff found 2 different rows.\n+ (20000, 942013020)\n- (20000, 942013021)\n[16:55:54] INFO - . . Diffing segment 1/4 of size 925929 and 925929\n[16:55:55] INFO - . . Diffing segment 2/4 of size 925929 and 925929\n[16:55:55] INFO - . . Diffing segment 3/4 of size 1 and 1\n[16:55:56] INFO - . Diffing segment 1/4 of size 2777788 and 2777788\n[16:55:58] INFO - . Diffing segment 2/4 of size 2777788 and 2777788\n[16:55:59] INFO - . Diffing segment 3/4 of size 1 and 1\n[16:56:00] INFO - Diffing segment 1/4 of size 8333365 and 8333365\n[16:56:06] INFO - Diffing segment 2/4 of size 8333365 and 8333365\n[16:56:11] INFO - Diffing segment 3/4 of size 1 and 1\n[16:56:11] INFO - Duration: 53.51 seconds.\n```\n\n\n# How to install\n\nRequires Python 3.7+ with pip.\n\n    poetry build --format wheel\n    pip install "dist/data_diff-0.0.2-py3-none-any.whl[mysql,pgsql]"\n\n# How to use\n\nUsage: `data_diff DB1_URI TABLE1_NAME DB2_URI TABLE2_NAME [OPTIONS]`\n\nOptions:\n\n  - `--help` - Show help message and exit.\n  - `-k` or `--key_column` - Name of the primary key column\n  - `-c` or `--columns` - List of names of extra columns to compare\n  - `-l` or `--limit` - Maximum number of differences to find (limits maximum bandwidth and runtime)\n  - `-s` or `--stats` - Print stats instead of a detailed diff\n  - `-d` or `--debug` - Print debug info\n  - `-v` or `--verbose` - Print extra info\n  - `--bisection-factor` - Segments per iteration. When set to 2, it performs binary search.\n  - `--bisection-threshold` - Minimal bisection threshold. i.e. maximum size of pages to diff locally.\n\n## Tips for performance\n\nIt\'s highly recommended that all involved columns are indexed.\n\n# License\n\nTBD',
    'author': 'Erez Shinnan',
    'author_email': 'erezshin@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/datafold/data-diff',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Provides classes for performing a table diff
"""

from typing import List, Tuple, Iterator, Literal
import logging

from runtype import dataclass

from .sql import Select, Checksum, Compare, DbPath, DbKey, Count, Enum, TableName, In, Value
from .database import Database

logger = logging.getLogger('diff_tables')

def safezip(*args):
    "zip but makes sure all sequences are the same length"
    assert len(set(map(len, args))) == 1
    return zip(*args)


@dataclass(frozen=False)
class TableSegment:
    database: Database
    table_path: DbPath
    key_column: str
    extra_columns: Tuple[str, ...]
    start: DbKey = None
    end: DbKey = None

    _count: int = None
    _checksum: int = None

    def _make_range_pred(self):
        if self.start is not None:
            yield Compare('<=', str(self.start), self.key_column)
        if self.end is not None:
            yield Compare('<', self.key_column, str(self.end))

    def _make_select(self, *, table=None, columns=None, where=None, group_by=None, order_by=None):
        if columns is None:
            columns = [self.key_column]
        where = list(self._make_range_pred()) + ([] if where is None else [where])
        order_by = None if order_by is None else [order_by]
        return Select(table=table or TableName(self.table_path), where=where, columns=columns, group_by=group_by, order_by=order_by)

    def get_values(self) -> list:
        "Download all the relevant values of the segment from the database"
        select = self._make_select(columns=self._relevant_columns)
        return self.database.query(select, List[Tuple])

    def choose_checkpoints(self, count: int) -> List[DbKey]:
        "Suggests a bunch of evenly-spaced checkpoints to split by"
        ratio = int(self.count / count)
        assert ratio > 1
        skip = f'mod(idx, {ratio}) = 0'
        select = self._make_select(table=Enum(self.table_path, order_by=self.key_column), where=skip)
        return self.database.query(select, List[int])

    def find_checkpoints(self, checkpoints: List[DbKey]) -> List[DbKey]:
        "Takes a list of potential checkpoints and returns those that exist"
        where = In(self.key_column, checkpoints)
        return self.database.query(self._make_select(where=where), List[int])

    def segment_by_checkpoints(self, checkpoints: List[DbKey]) -> List['TableSegment']:
        "Split the current TableSegment to a bunch of smaller ones, separate by the given checkpoints"

        if self.start and self.end:
            assert all(self.start <= c < self.end for c in checkpoints)
        checkpoints.sort()

        # Calculate sub-segments
        positions = [self.start] + checkpoints + [self.end]
        ranges = list(zip( positions[:-1], positions[1:] ))
        
        # Create table segments
        tables = [self.new(start=s, end=e) for s, e in ranges]

        return tables

        ## Calculate checksums in one go, to prevent repetitive individual calls
        # selects = [t._make_select(columns=[Checksum(self._relevant_columns)]) for t in tables]
        # res = self.database.query(Select(columns=selects), list)
        # checksums ,= res
        # assert len(checksums) == len(checkpoints) + 1
        # return [t.new(_checksum=checksum) for t, checksum in safezip(tables, checksums)]

    def new(self, _count=None, _checksum=None, **kwargs) -> 'TableSegment':
        """Using new() creates a copy of the instance using 'replace()', and makes sure the cache is reset"""
        return self.replace(_count=None, _checksum=None, **kwargs)

    @property
    def count(self) -> int:
        if self._count is None:
            self._count = self.database.query(self._make_select(columns=[Count()]), int)
        return self._count

    @property
    def _relevant_columns(self) -> List[str]:
        return [self.key_column] + list(self.extra_columns)

    @property
    def checksum(self) -> int:
        if self._checksum is None:
            self._checksum = self.database.query(self._make_select(columns=[Checksum(self._relevant_columns)]), int) or 0
        return self._checksum


def diff_sets(a: set, b: set) -> iter:
    s1 = set(a)
    s2 = set(b)
    for i in s1-s2:
        yield '+', i
    for i in s2-s1:
        yield '-', i

DiffResult = Iterator[Tuple[Literal['+', '-'], tuple]]

@dataclass
class TableDiffer:
    """Finds the diff between two SQL tables

    The algorithm uses hashing to quickly check if the tables are different, and then applies a
    bisection search recursively to find the differences efficiently.

    Works best for comparing tables that are mostly the name, with minor discrepencies.
    """

    bisection_factor: int = 32             # Into how many segments to bisect per iteration
    bisection_threshold: int = 1024**2   # When should we stop bisecting and compare locally (in row count)
    debug: bool = False

    def diff_tables(self, table1: TableSegment, table2: TableSegment) -> DiffResult:
        """Diff the given tables.

        Returned value is an iterator that yield pair-tuples, representing the diff. Items can be either
            ('+', columns) for items in table1 but not in table2
            ('-', columns) for items in table2 but not in table1
            Where `columns` is a tuple of values for the involved columns, i.e. (id, ...extra)
        """
        if self.bisection_factor >= self.bisection_threshold:
            raise ValueError("Incorrect param values")
        if self.bisection_factor < 2:
            raise ValueError("Must have at least two segments per iteration")

        logger.info(f'Diffing tables of size {table1.count} and {table2.count} | segments: {self.bisection_factor}, bisection threshold: {self.bisection_threshold}.')

        if table1.checksum == table2.checksum:
            return []   # No differences

        return self._diff_tables(table1, table2)

    def _diff_tables(self, table1, table2, level=0):
        count1 = table1.count
        count2 = table2.count

        # If count is below the threshold, just download and compare the columns locally
        # This saves time, as bisection speed is limited by ping and query performance.
        if count1 < self.bisection_threshold and count2 < self.bisection_threshold:
            rows1 = table1.get_values()
            rows2 = table2.get_values()
            diff = list(diff_sets(rows1, rows2))
            logger.info('. '*level + f'Diff found {len(diff)} different rows.')
            yield from diff
            return

        # Find mutual checkpoints between the two tables
        checkpoints = table1.choose_checkpoints(self.bisection_factor-1)
        assert checkpoints
        mutual_checkpoints = table2.find_checkpoints([Value(c) for c in checkpoints])
        mutual_checkpoints = list(set(mutual_checkpoints))        # Duplicate values are a problem!
        logger.debug('. '*level + f'Found {len(mutual_checkpoints)} mutual checkpoints (out of {len(checkpoints)}).')
        if not mutual_checkpoints:
            raise Exception("Tables are too different.")

        # Create new instances of TableSegment between each checkpoint
        segmented1 = table1.segment_by_checkpoints(mutual_checkpoints)
        segmented2 = table2.segment_by_checkpoints(mutual_checkpoints)
        if self.debug:
            logger.debug("Performing sanity tests for chosen segments (assert sum of fragments == whole)")
            assert count1 == sum(s.count for s in segmented1)
            assert count2 == sum(s.count for s in segmented2)

        # Compare each pair of corresponding segments between table1 and table2
        for i, (t1, t2) in enumerate(safezip(segmented1, segmented2)):
            logger.info('. '*level + f'Diffing segment {i+1}/{len(segmented1)} of size {t1.count} and {t2.count}')
            if t1.checksum != t2.checksum:
                # Apply recursively
                yield from self._diff_tables(t1, t2, level+1)
